@@ -1,6 +1,6 @@
 ;;; exwm-randr.el --- RandR Module for EXWM  -*- lexical-binding: t -*-
 
-;; Copyright (C) 2015-2016 Free Software Foundation, Inc.
+;; Copyright (C) 2015-2018 Free Software Foundation, Inc.
 
 ;; Author: Chris Feng <chris.w.feng@gmail.com>
 
@@ -21,9 +21,10 @@
 
 ;;; Commentary:
 
-;; This module adds RandR support for EXWM. Currently it requires external
-;; tools such as xrandr(1) to properly configure RandR first. This dependency
-;; may be removed in the future, but more work is needed before that.
+;; This module adds RandR support for EXWM.  Currently it requires external
+;; tools such as xrandr(1) to properly configure RandR first.  This
+;; dependency may be removed in the future, but more work is needed before
+;; that.
 
 ;; To use this module, load, enable it and configure
 ;; `exwm-randr-workspace-output-plist' and `exwm-randr-screen-change-hook'
@@ -49,27 +50,53 @@
 (require 'xcb-randr)
 (require 'exwm-core)
 
-(defvar exwm-randr-workspace-output-plist nil)
+(defgroup exwm-randr nil
+  "RandR."
+  :version "25.3"
+  :group 'exwm)
 
-(defvar exwm-randr-refresh-hook nil
-  "Normal hook run when the RandR module just refreshed.")
+(defcustom exwm-randr-refresh-hook nil
+  "Normal hook run when the RandR module just refreshed."
+  :type 'hook)
+
+(defcustom exwm-randr-screen-change-hook nil
+  "Normal hook run when screen changes."
+  :type 'hook)
+
+(defcustom exwm-randr-workspace-output-plist nil
+  "Plist mapping workspace to output.
+
+If an output is not available, the workspaces mapped to it are displayed on
+the primary output until it becomes available.  Unspecified workspaces are
+all mapped to the primary output.  For example, with the following value
+workspace other than 1 and 3 would always be displayed on the primary output
+where workspace 1 and 3 would be displayed on their corresponding output
+whenever the outputs are available.
+
+  '(1 \"HDMI-1\" 3 \"DP-1\")
+
+The outputs available can be identified by running the 'xrandr' utility with
+the first one in result being the primary output."
+  :type '(plist :key-type integer :value-type string))
 
 (defvar exwm-workspace--fullscreen-frame-count)
 (defvar exwm-workspace--list)
-
 (declare-function exwm-workspace--count "exwm-workspace.el")
-(declare-function exwm-workspace--set-fullscreen "exwm-workspace.el" (frame))
-(declare-function exwm-workspace--update-workareas "exwm-workspace.el" ())
-(declare-function exwm-workspace--show-minibuffer "exwm-workspace.el" ())
+(declare-function exwm-workspace--set-active "exwm-workspace.el"
+                  (frame active))
 (declare-function exwm-workspace--set-desktop-geometry "exwm-workspace.el" ())
+(declare-function exwm-workspace--set-fullscreen "exwm-workspace.el" (frame))
+(declare-function exwm-workspace--show-minibuffer "exwm-workspace.el" ())
+(declare-function exwm-workspace--update-workareas "exwm-workspace.el" ())
 
 (defun exwm-randr--refresh ()
   "Refresh workspaces according to the updated RandR info."
-  (let (output-name geometry output-plist default-geometry)
+  (let (output-name geometry output-plist primary-output default-geometry
+                    container-output-alist container-frame-alist)
     ;; Query all outputs
     (with-slots (config-timestamp outputs)
         (xcb:+request-unchecked+reply exwm--connection
-            (make-instance 'xcb:randr:GetScreenResources
+            (make-instance 'xcb:randr:GetScreenResourcesCurrent
                            :window exwm--root))
       (dolist (output outputs)
         (with-slots (crtc connection name)
@@ -91,8 +118,9 @@
                                             :x x :y y
                                             :width width :height height)
                     output-plist (plist-put output-plist output-name geometry))
-              (unless default-geometry ;assume the first output as primary
-                (setq default-geometry geometry)))))))
+              (unless primary-output
+                (setq primary-output output-name
+                      default-geometry geometry)))))))
     (exwm--log "(randr) outputs: %s" output-plist)
     (when output-plist
       (when exwm-workspace--fullscreen-frame-count
@@ -101,28 +129,52 @@
       (dotimes (i (exwm-workspace--count))
         (let* ((output (plist-get exwm-randr-workspace-output-plist i))
                (geometry (lax-plist-get output-plist output))
-               (frame (elt exwm-workspace--list i)))
+               (frame (elt exwm-workspace--list i))
+               (container (frame-parameter frame 'exwm-container)))
           (unless geometry
             (setq geometry default-geometry
-                  output nil))
+                  output primary-output))
+          (setq container-output-alist (nconc
+                                        `((,container . ,(intern output)))
+                                        container-output-alist)
+                container-frame-alist (nconc `((,container . ,frame))
+                                             container-frame-alist))
           (set-frame-parameter frame 'exwm-randr-output output)
           (set-frame-parameter frame 'exwm-geometry geometry)))
-      ;; Update workareas and set _NET_WORKAREA.
+      ;; Update workareas.
       (exwm-workspace--update-workareas)
       ;; Resize workspace.
       (dolist (f exwm-workspace--list)
         (exwm-workspace--set-fullscreen f))
+      (xcb:flush exwm--connection)
       ;; Raise the minibuffer if it's active.
       (when (and (active-minibuffer-window)
                  (exwm-workspace--minibuffer-own-frame-p))
         (exwm-workspace--show-minibuffer))
       ;; Set _NET_DESKTOP_GEOMETRY.
       (exwm-workspace--set-desktop-geometry)
+      ;; Update active/inactive workspaces.
+      (dolist (w exwm-workspace--list)
+        (exwm-workspace--set-active w nil))
+      (dolist (xwin
+               (reverse
+                (slot-value (xcb:+request-unchecked+reply exwm--connection
+                                (make-instance 'xcb:QueryTree
+                                               :window exwm--root))
+                            'children)))
+        (let ((output (cdr (assq xwin container-output-alist))))
+          (when output
+            (setq container-output-alist
+                  (rassq-delete-all output container-output-alist))
+            (exwm-workspace--set-active (cdr (assq xwin container-frame-alist))
+                                        t))))
       (xcb:flush exwm--connection)
       (run-hooks 'exwm-randr-refresh-hook))))
 
-(defvar exwm-randr-screen-change-hook nil
-  "Normal hook run when screen changes.")
+(defun exwm-randr--on-ScreenChangeNotify (_data _synthetic)
+  (exwm--log "(RandR) ScreenChangeNotify")
+  (run-hooks 'exwm-randr-screen-change-hook)
+  (exwm-randr--refresh))
 
 (defun exwm-randr--init ()
   "Initialize RandR extension and EXWM RandR module."
@@ -132,18 +184,15 @@
     (with-slots (major-version minor-version)
         (xcb:+request-unchecked+reply exwm--connection
             (make-instance 'xcb:randr:QueryVersion
-                           :major-version 1 :minor-version 2))
-      (if (or (/= major-version 1) (< minor-version 2))
+                           :major-version 1 :minor-version 3))
+      (if (or (/= major-version 1) (< minor-version 3))
           (error "[EXWM] The server only support RandR version up to %d.%d"
                  major-version minor-version)
         ;; External monitor(s) may already be connected.
         (run-hooks 'exwm-randr-screen-change-hook)
         (exwm-randr--refresh)
         (xcb:+event exwm--connection 'xcb:randr:ScreenChangeNotify
-                    (lambda (_data _synthetic)
-                      (exwm--log "(RandR) ScreenChangeNotify")
-                      (run-hooks 'exwm-randr-screen-change-hook)
-                      (exwm-randr--refresh)))
+                    #'exwm-randr--on-ScreenChangeNotify)
         ;; (xcb:+event exwm--connection 'xcb:randr:Notify
         ;;             (lambda (_data _synthetic)
         ;;               (exwm--log "(RandR) Notify")
@@ -160,7 +209,12 @@
                            ;;            xcb:randr:NotifyMask:CrtcChange))
                            ))
         (xcb:flush exwm--connection)
-        (add-hook 'exwm-workspace-list-change-hook #'exwm-randr--refresh)))))
+        (add-hook 'exwm-workspace-list-change-hook #'exwm-randr--refresh))))
+  ;; Prevent frame parameters introduced by this module from being
+  ;; saved/restored.
+  (dolist (i '(exwm-randr-output))
+    (unless (assq i frameset-filter-alist)
+      (push (cons i :never) frameset-filter-alist))))
 
 (defun exwm-randr--exit ()
   "Exit the RandR module."
