@@ -1,7 +1,7 @@
 ;;; exwm-systemtray.el --- System Tray Module for  -*- lexical-binding: t -*-
 ;;;                        EXWM
 
-;; Copyright (C) 2016 Free Software Foundation, Inc.
+;; Copyright (C) 2016-2018 Free Software Foundation, Inc.
 
 ;; Author: Chris Feng <chris.w.feng@gmail.com>
 
@@ -50,24 +50,38 @@
    (owner :initarg :owner :type xcb:WINDOW))      ;new slot
   :documentation "A systemtray client message.")
 
+(defgroup exwm-systemtray nil
+  "System tray."
+  :version "25.3"
+  :group 'exwm)
+
+(defcustom exwm-systemtray-height nil
+  "System tray height.
+
+You shall use the default value if using auto-hide minibuffer."
+  :type 'integer)
+
+(defcustom exwm-systemtray-icon-gap 2
+  "Gap between icons."
+  :type 'integer)
+
 ;; GTK icons require at least 16 pixels to show normally.
 (defconst exwm-systemtray--icon-min-size 16 "Minimum icon size.")
 
-(defvar exwm-systemtray-height nil "System tray height.
-
-You shall use the default value if using auto-hide minibuffer.")
-
-(defvar exwm-systemtray-icon-gap 2 "Gap between icons.")
-
 (defvar exwm-systemtray--connection nil "The X connection.")
-(defvar exwm-systemtray--list nil "The icon list.")
-(defvar exwm-systemtray--selection-owner-window nil
-  "The selection owner window.")
+
 (defvar exwm-systemtray--embedder nil "The embedder window.")
 
+(defvar exwm-systemtray--list nil "The icon list.")
+
+(defvar exwm-systemtray--selection-owner-window nil
+  "The selection owner window.")
+
 (defvar exwm-workspace--current)
-(declare-function exwm-workspace--current-height "exwm-workspace.el")
-(declare-function exwm-workspace--current-width  "exwm-workspace.el")
+(defvar exwm-workspace--minibuffer)
+(defvar exwm-workspace--workareas)
+(defvar exwm-workspace-current-index)
+(defvar xcb:Atom:_NET_SYSTEM_TRAY_S0)
 (declare-function exwm-workspace--minibuffer-own-frame-p "exwm-workspace.el")
 
 (defun exwm-systemtray--embed (icon)
@@ -118,7 +132,18 @@ You shall use the default value if using auto-hide minibuffer.")
                          :window icon
                          :value-mask xcb:CW:EventMask
                          :event-mask (logior xcb:EventMask:ResizeRedirect
+                                             xcb:EventMask:KeyPress
                                              xcb:EventMask:PropertyChange)))
+      ;; Grab all keys and forward them to Emacs frame.
+      (unless (exwm-workspace--minibuffer-own-frame-p)
+        (xcb:+request exwm-systemtray--connection
+            (make-instance 'xcb:GrabKey
+                           :owner-events 0
+                           :grab-window icon
+                           :modifiers xcb:ModMask:Any
+                           :key xcb:Grab:Any
+                           :pointer-mode xcb:GrabMode:Async
+                           :keyboard-mode xcb:GrabMode:Async)))
       (setq visible (slot-value info 'flags))
       (if visible
           (setq visible
@@ -178,13 +203,15 @@ You shall use the default value if using auto-hide minibuffer.")
         (setq x (+ x (slot-value (cdr pair) 'width)
                    exwm-systemtray-icon-gap))
         (setq map t)))
-    (xcb:+request exwm-systemtray--connection
-        (make-instance 'xcb:ConfigureWindow
-                       :window exwm-systemtray--embedder
-                       :value-mask (logior xcb:ConfigWindow:X
-                                           xcb:ConfigWindow:Width)
-                       :x (- (exwm-workspace--current-width) x)
-                       :width x))
+    (let ((workarea (elt exwm-workspace--workareas
+                         exwm-workspace-current-index)))
+      (xcb:+request exwm-systemtray--connection
+          (make-instance 'xcb:ConfigureWindow
+                         :window exwm-systemtray--embedder
+                         :value-mask (logior xcb:ConfigWindow:X
+                                             xcb:ConfigWindow:Width)
+                         :x (- (aref workarea 2) x)
+                         :width x)))
     (when map
       (xcb:+request exwm-systemtray--connection
           (make-instance 'xcb:MapWindow :window exwm-systemtray--embedder))))
@@ -276,6 +303,23 @@ You shall use the default value if using auto-hide minibuffer.")
               (t
                (exwm--log "(System Tray) Unknown opcode message: %s" obj)))))))
 
+(defun exwm-systemtray--on-KeyPress (data _synthetic)
+  "Forward all KeyPress events to Emacs frame."
+  ;; This function is only executed when there's no autohide minibuffer,
+  ;; a workspace frame has the input focus and the pointer is over a
+  ;; tray icon.
+  (let ((dest (frame-parameter (selected-frame) 'exwm-outer-id))
+        (obj (make-instance 'xcb:KeyPress)))
+    (xcb:unmarshal obj data)
+    (setf (slot-value obj 'event) dest)
+    (xcb:+request exwm-systemtray--connection
+        (make-instance 'xcb:SendEvent
+                       :propagate 0
+                       :destination dest
+                       :event-mask xcb:EventMask:NoEvent
+                       :event (xcb:marshal obj exwm-systemtray--connection))))
+  (xcb:flush exwm-systemtray--connection))
+
 (defun exwm-systemtray--on-workspace-switch ()
   "Reparent/Refresh the system tray in `exwm-workspace-switch-hook'."
   (unless (exwm-workspace--minibuffer-own-frame-p)
@@ -286,7 +330,7 @@ You shall use the default value if using auto-hide minibuffer.")
                                 (frame-parameter exwm-workspace--current
                                                  'window-id))
                        :x 0
-                       :y (- (exwm-workspace--current-height)
+                       :y (- (frame-pixel-height exwm-workspace--current)
                              exwm-systemtray-height))))
   (exwm-systemtray--refresh))
 
@@ -297,12 +341,12 @@ You shall use the default value if using auto-hide minibuffer.")
         (make-instance 'xcb:ConfigureWindow
                        :window exwm-systemtray--embedder
                        :value-mask xcb:ConfigWindow:Y
-                       :y (- (exwm-workspace--current-height)
+                       :y (- (frame-pixel-height exwm-workspace--current)
                              exwm-systemtray-height))))
   (exwm-systemtray--refresh))
 
-(defvar xcb:Atom:_NET_SYSTEM_TRAY_S0)
-(defvar exwm-workspace--minibuffer)
+(defalias 'exwm-systemtray--on-struts-update
+  #'exwm-systemtray--on-randr-refresh)
 
 (cl-defun exwm-systemtray--init ()
   "Initialize system tray module."
@@ -328,6 +372,7 @@ You shall use the default value if using auto-hide minibuffer.")
                          :selection xcb:Atom:_NET_SYSTEM_TRAY_S0))
     (when (/= owner xcb:Window:None)
       (xcb:disconnect exwm-systemtray--connection)
+      (setq exwm-systemtray--connection nil)
       (warn "[EXWM] Other system tray detected")
       (cl-return-from exwm-systemtray--init)))
   (let ((id (xcb:generate-id exwm-systemtray--connection)))
@@ -386,11 +431,14 @@ You shall use the default value if using auto-hide minibuffer.")
                     (- (line-pixel-height) exwm-systemtray-height)
                   ;; Vertically centered.
                   (/ (- (line-pixel-height) exwm-systemtray-height) 2)))
-      (setq frame exwm-workspace--current
-            ;; Bottom aligned.
-            y (- (exwm-workspace--current-height) exwm-systemtray-height)))
+      (let ((workarea (elt exwm-workspace--workareas
+                           exwm-workspace-current-index)))
+        (setq frame exwm-workspace--current
+              ;; Bottom aligned.
+              y (- (aref workarea 3) exwm-systemtray-height))))
     (setq parent (string-to-number (frame-parameter frame 'window-id))
-          depth (slot-value (xcb:+request-unchecked+reply exwm--connection
+          depth (slot-value (xcb:+request-unchecked+reply
+                                exwm-systemtray--connection
                                 (make-instance 'xcb:GetGeometry
                                                :drawable parent))
                             'depth))
@@ -425,14 +473,34 @@ You shall use the default value if using auto-hide minibuffer.")
               #'exwm-systemtray--on-PropertyNotify)
   (xcb:+event exwm-systemtray--connection 'xcb:ClientMessage
               #'exwm-systemtray--on-ClientMessage)
+  (unless (exwm-workspace--minibuffer-own-frame-p)
+    (xcb:+event exwm-systemtray--connection 'xcb:KeyPress
+                #'exwm-systemtray--on-KeyPress))
   ;; Add hook to move/reparent the embedder.
   (add-hook 'exwm-workspace-switch-hook #'exwm-systemtray--on-workspace-switch)
+  (add-hook 'exwm-workspace--update-workareas-hook
+            #'exwm-systemtray--on-struts-update)
   (when (boundp 'exwm-randr-refresh-hook)
-    (add-hook 'exwm-randr-refresh-hook #'exwm-systemtray--on-randr-refresh)))
+    (add-hook 'exwm-randr-refresh-hook #'exwm-systemtray--on-randr-refresh))
+  ;; The struts can be updated already.
+  (when exwm-workspace--workareas
+    (exwm-systemtray--on-struts-update)))
 
 (defun exwm-systemtray--exit ()
   "Exit the systemtray module."
   (when exwm-systemtray--connection
+    ;; Hide & reparent out the embedder before disconnection to prevent
+    ;; embedded icons from being reparented to an Emacs frame (which is the
+    ;; parent of the embedder).
+    (xcb:+request exwm-systemtray--connection
+        (make-instance 'xcb:UnmapWindow
+                       :window exwm-systemtray--embedder))
+    (xcb:+request exwm-systemtray--connection
+        (make-instance 'xcb:ReparentWindow
+                       :window exwm-systemtray--embedder
+                       :parent exwm--root
+                       :x 0
+                       :y 0))
     (xcb:disconnect exwm-systemtray--connection)
     (setq exwm-systemtray--connection nil
           exwm-systemtray--list nil
@@ -440,6 +508,8 @@ You shall use the default value if using auto-hide minibuffer.")
           exwm-systemtray--embedder nil)
     (remove-hook 'exwm-workspace-switch-hook
                  #'exwm-systemtray--on-workspace-switch)
+    (remove-hook 'exwm-workspace--update-workareas-hook
+                 #'exwm-systemtray--on-struts-update)
     (when (boundp 'exwm-randr-refresh-hook)
       (remove-hook 'exwm-randr-refresh-hook
                    #'exwm-systemtray--on-randr-refresh))))
@@ -453,4 +523,4 @@ You shall use the default value if using auto-hide minibuffer.")
 
 (provide 'exwm-systemtray)
 
-;; exwm-systemtray.el ends here
+;;; exwm-systemtray.el ends here
