@@ -1,11 +1,11 @@
 ;;; exwm.el --- Emacs X Window Manager  -*- lexical-binding: t -*-
 
-;; Copyright (C) 2015-2016 Free Software Foundation, Inc.
+;; Copyright (C) 2015-2018 Free Software Foundation, Inc.
 
 ;; Author: Chris Feng <chris.w.feng@gmail.com>
 ;; Maintainer: Chris Feng <chris.w.feng@gmail.com>
-;; Version: 0.9
-;; Package-Requires: ((xelb "0.11"))
+;; Version: 0.19
+;; Package-Requires: ((xelb "0.15"))
 ;; Keywords: unix
 ;; URL: https://github.com/ch11ng/exwm
 
@@ -36,7 +36,7 @@
 ;; + Dynamic workspace support
 ;; + ICCCM/EWMH compliance
 ;; + (Optional) RandR (multi-monitor) support
-;; + (Optional) Builtin system tray
+;; + (Optional) Built-in system tray
 
 ;; Installation & configuration
 ;; ----------------------------
@@ -72,13 +72,44 @@
 (require 'exwm-manage)
 (require 'exwm-input)
 
-;;;###autoload
+(defgroup exwm nil
+  "Emacs X Window Manager."
+  :tag "EXWM"
+  :version "25.3"
+  :group 'applications
+  :prefix "exwm-")
+
+(defcustom exwm-init-hook nil
+  "Normal hook run when EXWM has just finished initialization."
+  :type 'hook)
+
+(defcustom exwm-exit-hook nil
+  "Normal hook run just before EXWM exits."
+  :type 'hook)
+
+(defcustom exwm-update-class-hook nil
+  "Normal hook run when window class is updated."
+  :type 'hook)
+
+(defcustom exwm-update-title-hook nil
+  "Normal hook run when window title is updated."
+  :type 'hook)
+
+(defcustom exwm-blocking-subrs '(x-file-dialog x-popup-dialog x-select-font)
+  "Subrs (primitives) that would normally block EXWM."
+  :type '(repeat function))
+
+(defconst exwm--server-name "server-exwm"
+  "Name of the subordinate Emacs server.")
+
+(defvar exwm--server-process nil "Process of the subordinate Emacs server.")
+
 (defun exwm-reset ()
-  "Reset window to standard state: non-fullscreen, line-mode."
+  "Reset the state of the selected window (non-fullscreen, line-mode, etc)."
   (interactive)
   (with-current-buffer (window-buffer)
-    (when (eq major-mode 'exwm-mode)
-      (when (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN exwm--ewmh-state)
+    (when (derived-mode-p 'exwm-mode)
+      (when (exwm-layout--fullscreen-p)
         (exwm-layout-unset-fullscreen))
       ;; Force refresh
       (exwm-layout--refresh)
@@ -88,7 +119,7 @@
 (defun exwm-restart ()
   "Restart EXWM."
   (interactive)
-  (when (exwm-workspace--confirm-kill-emacs "[EXWM] Restart? " 'no-check)
+  (when (exwm--confirm-kill-emacs "[EXWM] Restart? " 'no-check)
     (let* ((attr (process-attributes (emacs-pid)))
            (args (cdr (assq 'args attr)))
            (ppid (cdr (assq 'ppid attr)))
@@ -113,6 +144,32 @@
           (while (= ?R (shell-command-on-region (point) (point) args))))
         (kill-emacs))))))
 
+(defun exwm--update-desktop (xwin)
+  "Update _NET_WM_DESKTOP."
+  (with-current-buffer (exwm--id->buffer xwin)
+    (let ((reply (xcb:+request-unchecked+reply exwm--connection
+                     (make-instance 'xcb:ewmh:get-_NET_WM_DESKTOP
+                                    :window xwin)))
+          desktop)
+      (when reply
+        (setq desktop (slot-value reply 'value))
+        (cond
+         ((eq desktop 4294967295.)
+          (unless (or (not exwm--floating-frame)
+                      (eq exwm--frame exwm-workspace--current)
+                      (and exwm--desktop
+                           (= desktop exwm--desktop)))
+            (exwm-layout--show xwin (frame-root-window exwm--floating-frame)))
+          (setq exwm--desktop desktop))
+         ((and desktop
+               (< desktop (exwm-workspace--count))
+               (if exwm--desktop
+                   (/= desktop exwm--desktop)
+                 (/= desktop (exwm-workspace--position exwm--frame))))
+          (exwm-workspace-move-window desktop xwin))
+         (t
+          (exwm-workspace--set-desktop xwin)))))))
+
 (defun exwm--update-window-type (id &optional force)
   "Update _NET_WM_WINDOW_TYPE."
   (with-current-buffer (exwm--id->buffer id)
@@ -123,9 +180,6 @@
         (when reply                     ;nil when destroyed
           (setq exwm-window-type (append (slot-value reply 'value) nil)))))))
 
-(defvar exwm-update-class-hook nil
-  "Normal hook run when window class is updated.")
-
 (defun exwm--update-class (id &optional force)
   "Update WM_CLASS."
   (with-current-buffer (exwm--id->buffer id)
@@ -138,9 +192,6 @@
           (when (and exwm-instance-name exwm-class-name)
             (run-hooks 'exwm-update-class-hook)))))))
 
-(defvar exwm-update-title-hook nil
-  "Normal hook run when window title is updated.")
-
 (defun exwm--update-utf8-title (id &optional force)
   "Update _NET_WM_NAME."
   (with-current-buffer (exwm--id->buffer id)
@@ -272,7 +323,7 @@
             (setcdr pair struts)
           (push (cons id struts) exwm-workspace--id-struts-alist))
         (exwm-workspace--update-struts))
-      ;; Update workareas and set _NET_WORKAREA.
+      ;; Update workareas.
       (exwm-workspace--update-workareas)
       ;; Update workspaces.
       (dolist (f exwm-workspace--list)
@@ -291,7 +342,7 @@
           (setcdr pair struts)
         (push (cons id struts) exwm-workspace--id-struts-alist))
       (exwm-workspace--update-struts))
-    ;; Update workareas and set _NET_WORKAREA.
+    ;; Update workareas.
     (exwm-workspace--update-workareas)
     ;; Update workspaces.
     (dolist (f exwm-workspace--list)
@@ -398,19 +449,20 @@
      ;; _NET_REQUEST_FRAME_EXTENTS
      ((= type xcb:Atom:_NET_REQUEST_FRAME_EXTENTS)
       (let ((buffer (exwm--id->buffer id))
-            left right top btm)
+            top btm)
         (if (or (not buffer)
                 (not (buffer-local-value 'exwm--floating-frame buffer)))
-            (setq left 0 right 0 top 0 btm 0)
-          (setq left exwm-floating-border-width
-                right exwm-floating-border-width
-                top (+ exwm-floating-border-width (window-header-line-height))
-                btm (+ exwm-floating-border-width
-                       (window-mode-line-height))))
+            (setq top 0
+                  btm 0)
+          (setq top (window-header-line-height)
+                btm (window-mode-line-height)))
         (xcb:+request exwm--connection
             (make-instance 'xcb:ewmh:set-_NET_FRAME_EXTENTS
-                           :window id :left left :right right
-                           :top top :bottom btm)))
+                           :window id
+                           :left 0
+                           :right 0
+                           :top top
+                           :bottom btm)))
       (xcb:flush exwm--connection))
      ;; _NET_WM_DESKTOP.
      ((= type xcb:Atom:_NET_WM_DESKTOP)
@@ -427,47 +479,26 @@
         (when (and (not buffer)
                    (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN props)
                    (= action xcb:ewmh:_NET_WM_STATE_ADD))
-          (dolist (f exwm-workspace--list)
-            (when (equal (frame-parameter f 'exwm-outer-id) id)
-              (exwm-workspace--set-fullscreen f)
-              (xcb:+request
-                  exwm--connection
-                  (make-instance 'xcb:ewmh:set-_NET_WM_STATE
-                                 :window id
-                                 :data (vector
-                                        xcb:Atom:_NET_WM_STATE_FULLSCREEN)))
-              (xcb:flush exwm--connection))))
+          (xcb:+request
+              exwm--connection
+              (make-instance 'xcb:ewmh:set-_NET_WM_STATE
+                             :window id
+                             :data (vector xcb:Atom:_NET_WM_STATE_FULLSCREEN)))
+          (xcb:flush exwm--connection))
         (when buffer                    ;ensure it's managed
           (with-current-buffer buffer
-            ;; _NET_WM_STATE_MODAL
-            (when (memq xcb:Atom:_NET_WM_STATE_MODAL props)
-              (cond ((= action xcb:ewmh:_NET_WM_STATE_ADD)
-                     (unless exwm--floating-frame
-                       (exwm-floating--set-floating id))
-                     (push xcb:Atom:_NET_WM_STATE_MODAL props-new))
-                    ((= action xcb:ewmh:_NET_WM_STATE_REMOVE)
-                     (when exwm--floating-frame
-                       (exwm-floating--unset-floating id)))
-                    ((= action xcb:ewmh:_NET_WM_STATE_TOGGLE)
-                     (if exwm--floating-frame
-                         (exwm-floating--unset-floating id)
-                       (exwm-floating--set-floating id)
-                       (push xcb:Atom:_NET_WM_STATE_MODAL props-new)))))
             ;; _NET_WM_STATE_FULLSCREEN
             (when (or (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN props)
                       (memq xcb:Atom:_NET_WM_STATE_ABOVE props))
               (cond ((= action xcb:ewmh:_NET_WM_STATE_ADD)
-                     (unless (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN
-                                   exwm--ewmh-state)
+                     (unless (exwm-layout--fullscreen-p)
                        (exwm-layout-set-fullscreen id))
                      (push xcb:Atom:_NET_WM_STATE_FULLSCREEN props-new))
                     ((= action xcb:ewmh:_NET_WM_STATE_REMOVE)
-                     (when (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN
-                                 exwm--ewmh-state)
+                     (when (exwm-layout--fullscreen-p)
                        (exwm-layout-unset-fullscreen id)))
                     ((= action xcb:ewmh:_NET_WM_STATE_TOGGLE)
-                     (if (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN
-                               exwm--ewmh-state)
+                     (if (exwm-layout--fullscreen-p)
                          (exwm-layout-unset-fullscreen id)
                        (exwm-layout-set-fullscreen id)
                        (push xcb:Atom:_NET_WM_STATE_FULLSCREEN props-new)))))
@@ -498,12 +529,26 @@
             (bury-buffer)))))
      (t (exwm--log "Unhandled client message: %s" obj)))))
 
+(defun exwm--on-SelectionClear (data _synthetic)
+  "Handle SelectionClear events."
+  (exwm--log "SelectionClear")
+  (let ((obj (make-instance 'xcb:SelectionClear))
+        owner selection)
+    (xcb:unmarshal obj data)
+    (setq owner (slot-value obj 'owner)
+          selection (slot-value obj 'selection))
+    (when (and (eq owner exwm--wmsn-window)
+               (eq selection xcb:Atom:WM_S0))
+      (exwm-exit))))
+
 (defun exwm--init-icccm-ewmh ()
   "Initialize ICCCM/EWMH support."
   ;; Handle PropertyNotify event
   (xcb:+event exwm--connection 'xcb:PropertyNotify #'exwm--on-PropertyNotify)
   ;; Handle relevant client messages
   (xcb:+event exwm--connection 'xcb:ClientMessage #'exwm--on-ClientMessage)
+  ;; Handle SelectionClear
+  (xcb:+event exwm--connection 'xcb:SelectionClear #'exwm--on-SelectionClear)
   ;; Set _NET_SUPPORTED
   (xcb:+request exwm--connection
       (make-instance 'xcb:ewmh:set-_NET_SUPPORTED
@@ -519,9 +564,9 @@
                             xcb:Atom:_NET_CURRENT_DESKTOP
                             ;; xcb:Atom:_NET_DESKTOP_NAMES
                             xcb:Atom:_NET_ACTIVE_WINDOW
-                            xcb:Atom:_NET_WORKAREA
+                            ;; xcb:Atom:_NET_WORKAREA
                             xcb:Atom:_NET_SUPPORTING_WM_CHECK
-                            xcb:Atom:_NET_VIRTUAL_ROOTS
+                            ;; xcb:Atom:_NET_VIRTUAL_ROOTS
                             ;; xcb:Atom:_NET_DESKTOP_LAYOUT
                             ;; xcb:Atom:_NET_SHOWING_DESKTOP
 
@@ -556,7 +601,7 @@
                             xcb:Atom:_NET_WM_WINDOW_TYPE_NORMAL
                             ;;
                             xcb:Atom:_NET_WM_STATE
-                            xcb:Atom:_NET_WM_STATE_MODAL
+                            ;; xcb:Atom:_NET_WM_STATE_MODAL
                             ;; xcb:Atom:_NET_WM_STATE_STICKY
                             ;; xcb:Atom:_NET_WM_STATE_MAXIMIZED_VERT
                             ;; xcb:Atom:_NET_WM_STATE_MAXIMIZED_HORZ
@@ -605,13 +650,14 @@
                             xcb:Atom:_NET_WM_FULL_PLACEMENT)))
   ;; Create a child window for setting _NET_SUPPORTING_WM_CHECK
   (let ((new-id (xcb:generate-id exwm--connection)))
+    (setq exwm--guide-window new-id)
     (xcb:+request exwm--connection
         (make-instance 'xcb:CreateWindow
                        :depth 0
                        :wid new-id
                        :parent exwm--root
-                       :x 0
-                       :y 0
+                       :x -1
+                       :y -1
                        :width 1
                        :height 1
                        :border-width 0
@@ -635,63 +681,120 @@
                      :data [0 0]))
   (xcb:flush exwm--connection))
 
-(defun exwm--exit-icccm-ewmh ()
-  "Remove ICCCM/EWMH properties."
-  (dolist (p (list
-              xcb:Atom:_NET_WM_NAME
-              xcb:Atom:_NET_SUPPORTED
-              xcb:Atom:_NET_CLIENT_LIST
-              xcb:Atom:_NET_CLIENT_LIST_STACKING
-              xcb:Atom:_NET_NUMBER_OF_DESKTOPS
-              xcb:Atom:_NET_DESKTOP_GEOMETRY
-              xcb:Atom:_NET_DESKTOP_VIEWPORT
-              xcb:Atom:_NET_CURRENT_DESKTOP
-              xcb:Atom:_NET_ACTIVE_WINDOW
-              xcb:Atom:_NET_WORKAREA
-              xcb:Atom:_NET_SUPPORTING_WM_CHECK
-              xcb:Atom:_NET_VIRTUAL_ROOTS
-              ;; TODO: Keep this list synchronized with that in
-              ;;       `exwm--init-icccm-ewmh'.
-              ))
-    (xcb:+request exwm--connection
-        (make-instance 'xcb:DeleteProperty
-                       :window exwm--root
-                       :property p))
-    (xcb:flush exwm--connection)))
+(defun exwm--wmsn-acquire (replace)
+  "Acquire the WM_Sn selection.
 
-(defvar exwm-init-hook nil
-  "Normal hook run when EXWM has just finished initialization.")
+REPLACE specifies what to do in case there already is a window
+manager.  If t, replace it, if nil, abort and ask the user if `ask'."
+  (with-slots (owner)
+      (xcb:+request-unchecked+reply exwm--connection
+          (make-instance 'xcb:GetSelectionOwner
+                         :selection xcb:Atom:WM_S0))
+    (when (/= owner xcb:Window:None)
+      (when (eq replace 'ask)
+        (setq replace (yes-or-no-p "Replace existing window manager? ")))
+      (when (not replace)
+        (error "Other window manager detected")))
+    (let ((new-owner (xcb:generate-id exwm--connection)))
+      (xcb:+request exwm--connection
+          (make-instance 'xcb:CreateWindow
+                         :depth 0
+                         :wid new-owner
+                         :parent exwm--root
+                         :x -1
+                         :y -1
+                         :width 1
+                         :height 1
+                         :border-width 0
+                         :class xcb:WindowClass:CopyFromParent
+                         :visual 0
+                         :value-mask 0
+                         :override-redirect 0))
+      (xcb:+request exwm--connection
+          (make-instance 'xcb:ewmh:set-_NET_WM_NAME
+                         :window new-owner :data "EXWM selection owner"))
+      (xcb:+request-checked+request-check exwm--connection
+          (make-instance 'xcb:SetSelectionOwner
+                         :selection xcb:Atom:WM_S0
+                         :owner new-owner
+                         :time xcb:Time:CurrentTime))
+      (with-slots (owner)
+          (xcb:+request-unchecked+reply exwm--connection
+              (make-instance 'xcb:GetSelectionOwner
+                             :selection xcb:Atom:WM_S0))
+        (unless (eq owner new-owner)
+          (error "Could not acquire ownership of WM selection")))
+      ;; Wait for the other window manager to terminate.
+      (when (/= owner xcb:Window:None)
+        (let (reply)
+          (cl-dotimes (i exwm--wmsn-acquire-timeout)
+            (setq reply (xcb:+request-unchecked+reply exwm--connection
+                            (make-instance 'xcb:GetGeometry :drawable owner)))
+            (when (not reply)
+              (cl-return))
+            (message "Waiting for other window manager to quit... %ds" i)
+            (sleep-for 1))
+          (when reply
+            (error "Other window manager did not release selection in time"))))
+      ;; announce
+      (let* ((cmd (make-instance 'xcb:ClientMessageData
+                                 :data32 (vector xcb:Time:CurrentTime
+                                                 xcb:Atom:WM_S0
+                                                 new-owner
+                                                 0
+                                                 0)))
+             (cm (make-instance 'xcb:ClientMessage
+                                               :window exwm--root
+                                               :format 32
+                                               :type xcb:Atom:MANAGER
+                                               :data cmd))
+             (se (make-instance 'xcb:SendEvent
+                         :propagate 0
+                         :destination exwm--root
+                         :event-mask xcb:EventMask:NoEvent
+                         :event (xcb:marshal cm exwm--connection))))
+        (xcb:+request exwm--connection se))
+      (setq exwm--wmsn-window new-owner))))
 
-(defun exwm-init (&optional frame)
+;;;###autoload
+(cl-defun exwm-init (&optional frame)
   "Initialize EXWM."
+  (interactive)
   (if frame
       ;; The frame might not be selected if it's created by emacslicnet.
       (select-frame-set-input-focus frame)
     (setq frame (selected-frame)))
-  (if (not (eq 'x (framep frame)))
-      (exwm--log "Not running under X environment")
-    (unless exwm--connection
-      (exwm-enable 'undo)               ;never initialize again
-      (setq exwm--connection (xcb:connect))
-      (set-process-query-on-exit-flag (slot-value exwm--connection 'process)
-                                      nil) ;prevent query message on exit
-      (setq exwm--root
-            (slot-value (car (slot-value
-                              (xcb:get-setup exwm--connection) 'roots))
-                        'root))
-      (if (xcb:+request-checked+request-check exwm--connection
-              (make-instance 'xcb:ChangeWindowAttributes
-                             :window exwm--root :value-mask xcb:CW:EventMask
-                             :event-mask xcb:EventMask:SubstructureRedirect))
-          ;; Other window manager is running
-          (progn (xcb:disconnect exwm--connection)
-                 (setq exwm--connection nil)
-                 (exwm--log "Other window manager detected"))
-        ;; Disable some features not working well with EXWM
-        (setq use-dialog-box nil)
+  (when (not (eq 'x (framep frame)))
+    (message "[EXWM] Not running under X environment")
+    (cl-return-from exwm-init))
+  (when exwm--connection
+    (exwm--log "EXWM already running")
+    (cl-return-from exwm-init))
+  (condition-case err
+      (progn
+        (exwm-enable 'undo)               ;never initialize again
+        (setq exwm--connection (xcb:connect))
+        (set-process-query-on-exit-flag (slot-value exwm--connection 'process)
+                                        nil) ;prevent query message on exit
+        (setq exwm--root
+              (slot-value (car (slot-value
+                                (xcb:get-setup exwm--connection) 'roots))
+                          'root))
         ;; Initialize ICCCM/EWMH support
         (xcb:icccm:init exwm--connection t)
         (xcb:ewmh:init exwm--connection t)
+        ;; Try to register window manager selection.
+        (exwm--wmsn-acquire 'ask)
+        (when (xcb:+request-checked+request-check exwm--connection
+                  (make-instance 'xcb:ChangeWindowAttributes
+                                 :window exwm--root
+                                 :value-mask xcb:CW:EventMask
+                                 :event-mask
+                                 xcb:EventMask:SubstructureRedirect))
+          (error "Other window manager is running"))
+        ;; Disable some features not working well with EXWM
+        (setq use-dialog-box nil
+              confirm-kill-emacs #'exwm--confirm-kill-emacs)
         (exwm--lock)
         (exwm--init-icccm-ewmh)
         (exwm-layout--init)
@@ -701,25 +804,32 @@
         (exwm-input--init)
         (exwm--unlock)
         (exwm-workspace--post-init)
+        (exwm-input--post-init)
         ;; Manage existing windows
         (exwm-manage--scan)
-        (run-hooks 'exwm-init-hook)))))
+        (run-hooks 'exwm-init-hook))
+    ((quit error)
+     (exwm-exit)
+     ;; Rethrow error
+     (warn "[EXWM] EXWM fails to start (%s: %s)" (car err) (cdr err)))))
 
-(defvar exwm-exit-hook nil "Normal hook run just before EXWM exits.")
 
-(defun exwm--exit ()
+;;;###autoload
+(defun exwm-exit ()
   "Exit EXWM."
+  (interactive)
   (run-hooks 'exwm-exit-hook)
+  (setq confirm-kill-emacs nil)
   ;; Exit modules.
   (exwm-input--exit)
-  (exwm-workspace--exit)
   (exwm-manage--exit)
+  (exwm-workspace--exit)
   (exwm-floating--exit)
   (exwm-layout--exit)
-  (exwm--exit-icccm-ewmh))
-
-(defvar exwm-blocking-subrs '(x-file-dialog x-popup-dialog x-select-font)
-  "Subrs (primitives) that would normally block EXWM.")
+  (when exwm--connection
+    (xcb:flush exwm--connection)
+    (xcb:disconnect exwm--connection))
+  (setq exwm--connection nil))
 
 (defun exwm-enable (&optional undo)
   "Enable/Disable EXWM."
@@ -739,16 +849,14 @@
      ;; Ignore unrecognized command line arguments.  This can be helpful
      ;; when EXWM is launched by some session manager.
      (push #'vector command-line-functions)
-     (add-hook 'window-setup-hook #'exwm-init t)          ;for Emacs
-     (add-hook 'after-make-frame-functions #'exwm-init t) ;for Emacs Client
+     (add-hook (if (daemonp)
+                   'after-make-frame-functions ;emacsclient
+                 'window-setup-hook)           ;emacs
+               #'exwm-init t)
      (add-hook 'kill-emacs-hook #'exwm--server-stop)
      (dolist (i exwm-blocking-subrs)
        (advice-add i :around #'exwm--server-eval-at)))))
 
-(defconst exwm--server-name "server-exwm"
-  "Name of the subordinate Emacs server.")
-(defvar exwm--server-process nil "Process of the subordinate Emacs server.")
-
 (defun exwm--server-stop ()
   "Stop the subordinate Emacs server."
   (server-force-delete exwm--server-name)
@@ -795,12 +903,38 @@
                ;; For other types, return the value as-is.
                (t result))))))
 
-(define-obsolete-function-alias 'exwm-enable-ido-workaround 'exwm-config-ido
-  "25.1" "Enable workarounds for Ido.")
-
-(defun exwm-disable-ido-workaround ()
-  "This function does nothing actually."
-  (declare (obsolete nil "25.1")))
+(defun exwm--confirm-kill-emacs (prompt &optional force)
+  "Confirm before exiting Emacs."
+  (when (cond
+         ((and force (not (eq force 'no-check)))
+          ;; Force killing Emacs.
+          t)
+         ((or (eq force 'no-check) (not exwm--id-buffer-alist))
+          ;; Check if there's any unsaved file.
+          (pcase (catch 'break
+                   (let ((kill-emacs-query-functions
+                          (append kill-emacs-query-functions
+                                  (list (lambda ()
+                                          (throw 'break 'break))))))
+                     (save-buffers-kill-emacs)))
+            (`break (y-or-n-p prompt))
+            (x x)))
+         (t
+          (yes-or-no-p (format "[EXWM] %d window(s) will be destroyed.  %s"
+                               (length exwm--id-buffer-alist) prompt))))
+    ;; Run `kill-emacs-hook' (`server-force-stop' excluded) before Emacs
+    ;; frames are unmapped so that errors (if any) can be visible.
+    (if (memq #'server-force-stop kill-emacs-hook)
+        (progn
+          (setq kill-emacs-hook (delq #'server-force-stop kill-emacs-hook))
+          (run-hooks 'kill-emacs-hook)
+          (setq kill-emacs-hook (list #'server-force-stop)))
+      (run-hooks 'kill-emacs-hook)
+      (setq kill-emacs-hook nil))
+    ;; Exit each module, destroying all resources created by this connection.
+    (exwm-exit)
+    ;; Set the return value.
+    t))
 
 
 
